@@ -5,7 +5,7 @@ from skimage import draw
 from scipy.ndimage import filters, morphology
 import cv2
 import numpy as np
-from shapely.geometry import asPolygon, Polygon
+from shapely.geometry import Polygon
 from shapely.ops import unary_union
 
 from ocrd import Processor
@@ -41,6 +41,7 @@ from ocrd_validators.page_validator import (
     PageValidator
 )
 from .config import OCRD_TOOL
+from .project import join_polygons
 
 TOOL = 'ocrd-segment-repair'
 
@@ -60,6 +61,11 @@ class RepairSegmentation(Processor):
         Fix inconsistencies by shrinking segment polygons to their parents. Log
         errors that cannot be repaired automatically.
         
+        Next, if ``simplify`` is non-zero, then for each segment (top-level page or
+        recursive region, line, word, glyph), simplify the polygon points up to that
+        precision, while preserving its topology and parent-child consistency.
+        (This will usually reduce the number of points.)
+        
         \b
         Next, if ``plausibilize``, then for each segment (top-level page or recursive region)
         which contains any text regions, try to find all pairs of such regions in it that
@@ -96,6 +102,7 @@ class RepairSegmentation(Processor):
         assert_file_grp_cardinality(self.input_file_grp, 1)
         assert_file_grp_cardinality(self.output_file_grp, 1)
 
+        simplify = self.parameter['simplify']
         sanitize = self.parameter['sanitize']
         plausibilize = self.parameter['plausibilize']
         
@@ -165,12 +172,20 @@ class RepairSegmentation(Processor):
             # show remaining errors
             if not report.is_valid:
                 LOG.warning(report.to_xml())
+            # simplify
+            if simplify:
+                self.simplify_page(page, page_id)
             # delete/merge/split redundant text regions (or its text lines)
             if plausibilize:
                 self.plausibilize_page(page, page_id)
             # shrink/expand text regions to the hull of their text lines
             if sanitize:
-                self.sanitize_page(page, page_id)
+                page_image, page_coords, _ = self.workspace.image_from_page(
+                    page, page_id,
+                    feature_selector='binarized',
+                    feature_filter='clipped')
+                shrink_regions(page_image, page_coords, page, page_id,
+                               padding=self.parameter['sanitize_padding'])
             
             file_id = make_file_id(input_file, self.output_file_grp)
             self.workspace.add_file(
@@ -181,6 +196,25 @@ class RepairSegmentation(Processor):
                 local_filename=os.path.join(self.output_file_grp,
                                             file_id + '.xml'),
                 content=to_xml(pcgts))
+
+    def simplify_page(self, page, page_id):
+        if page.get_Border() is not None:
+            simplify(page.get_Border(), self.parameter['simplify'])
+        for region in page.get_AllRegions():
+            simplify(region, self.parameter['simplify'])
+            if isinstance(region, TextRegionType):
+                for line in region.get_TextLine():
+                    simplify(line, self.parameter['simplify'])
+                    for word in line.get_Word():
+                        simplify(word, self.parameter['simplify'])
+                        for glyph in word.get_Glyph():
+                            simplify(glyph, self.parameter['simplify'])
+                            ensure_consistent(glyph, at_parent=True)
+                        ensure_consistent(word, at_parent=True)
+                    ensure_consistent(line, at_parent=True)
+            ensure_consistent(region, at_parent=True)
+        if page.get_Border() is not None:
+            ensure_consistent(page)
     
     def plausibilize_page(self, page, page_id):
         ro = page.get_ReadingOrder()
@@ -252,90 +286,6 @@ class RepairSegmentation(Processor):
                                    marked_for_merging,
                                    marked_for_splitting)
     
-    def sanitize_page(self, page, page_id):
-        """Shrink each region outline to become the minimal convex hull of its constituent textlines."""
-        # FIXME: should probably be removed in favour of ocrd-segment-project entirely
-        LOG = getLogger('processor.RepairSegmentation')
-        regions = page.get_AllRegions(classes=['Text'])
-        page_image, page_coords, _ = self.workspace.image_from_page(
-            page, page_id)
-        for region in regions:
-            #LOG.info('Sanitizing region "%s"', region.id)
-            lines = region.get_TextLine()
-            if not lines:
-                LOG.warning('Page "%s" region "%s" contains no textlines', page_id, region.id)
-                continue
-            heights = []
-            tops = []
-            # get labels:
-            region_mask = np.zeros((page_image.height, page_image.width), dtype=np.uint8)
-            for line in lines:
-                line_polygon = coordinates_of_segment(line, page_image, page_coords)
-                line_xywh = xywh_from_polygon(line_polygon)
-                heights.append(line_xywh['h'])
-                tops.append(line_xywh['y'])
-                region_mask[draw.polygon(line_polygon[:, 1],
-                                         line_polygon[:, 0],
-                                         region_mask.shape)] = 1
-                region_mask[draw.polygon_perimeter(line_polygon[:, 1],
-                                                   line_polygon[:, 0],
-                                                   region_mask.shape)] = 1
-            # estimate scale:
-            heights = np.array(heights)
-            scale = int(np.max(heights))
-            tops = np.array(tops)
-            order = np.argsort(tops)
-            heights = heights[order]
-            tops = tops[order]
-            if len(lines) > 1:
-                # if interline spacing is larger than line height, use this
-                bottoms = tops + heights
-                deltas = tops[1:] - bottoms[:-1]
-                scale = max(scale, int(np.max(deltas)))
-            # close labels:
-            region_mask = np.pad(region_mask, scale) # protect edges
-            region_mask = np.array(morphology.binary_closing(region_mask, np.ones((scale, 1))), dtype=np.uint8)
-            region_mask = region_mask[scale:-scale, scale:-scale] # unprotect
-            # extend margins (to ensure simplified hull polygon is outside children):
-            region_mask = filters.maximum_filter(region_mask, 3) # 1px in each direction
-            # find outer contour (parts):
-            contours, _ = cv2.findContours(region_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
-            # determine areas of parts:
-            areas = [cv2.contourArea(contour) for contour in contours]
-            total_area = sum(areas)
-            if not total_area:
-                # ignore if too small
-                LOG.warning('Zero contour area in region "%s"', region.id)
-                continue
-            # pick contour and convert to absolute:
-            region_polygon = None
-            for i, contour in enumerate(contours):
-                area = areas[i]
-                if area / total_area < 0.1:
-                    LOG.warning('Ignoring contour %d too small (%d/%d) in region "%s"',
-                                i, area, total_area, region.id)
-                    continue
-                # simplify shape (until valid):
-                # can produce invalid (self-intersecting) polygons:
-                #polygon = cv2.approxPolyDP(contour, 2, False)[:, 0, ::] # already ordered x,y
-                polygon = contour[:, 0, ::] # already ordered x,y
-                polygon = Polygon(polygon).simplify(1)
-                polygon = make_valid(polygon)
-                polygon = polygon.exterior.coords[:-1] # keep open
-                if len(polygon) < 4:
-                    LOG.warning('Ignoring contour %d less than 4 points in region "%s"',
-                                i, region.id)
-                    continue
-                if region_polygon is not None:
-                    LOG.error('Skipping region "%s" due to non-contiguous contours',
-                              region.id)
-                    region_polygon = None
-                    break
-                region_polygon = coordinates_for_segment(polygon, page_image, page_coords)
-            if region_polygon is not None:
-                LOG.debug('Using new coordinates for region "%s"', region.id)
-                region.get_Coords().set_points(points_from_polygon(region_polygon))
-    
 def _compare_segments(seg1, seg2, poly1, poly2, marked_for_deletion, marked_for_merging, min_overlap, page_id):
     """Determine redundancies in a pair of regions/lines
     
@@ -425,12 +375,7 @@ def _merge_segments(seg, superseg, poly, superpoly, segpolys, reading_order):
     # show warnings when granularity is lost; but there might
     # be good reasons to do more here when we have better processors
     # and use-cases in the future
-    superpoly = superpoly.union(poly)
-    if superpoly.type == 'MultiPolygon':
-        superpoly = superpoly.convex_hull
-    if superpoly.minimum_clearance < 1.0:
-        superpoly = asPolygon(np.round(superpoly.exterior.coords))
-    superpoly = make_valid(superpoly)
+    superpoly = merge_poly(superpoly, poly)
     superpoly = superpoly.exterior.coords[:-1] # keep open
     superseg.get_Coords().set_points(points_from_polygon(superpoly))
     # FIXME should we merge/mix attributes and features?
@@ -522,7 +467,7 @@ def _plausibilize_segments(segpolys, rogroup, marked_for_deletion, marked_for_me
             if poly.type == 'MultiPolygon':
                 poly = poly.convex_hull
             if poly.minimum_clearance < 1.0:
-                poly = asPolygon(np.round(poly.exterior.coords))
+                poly = Polygon(np.round(poly.exterior.coords))
             poly = make_valid(poly)
             poly = poly.exterior.coords[:-1] # keep open
             seg.get_Coords().set_points(points_from_polygon(poly))
@@ -574,49 +519,181 @@ def page_get_reading_order(ro, rogroup):
             # recursive reading order element (un/ordered group):
             page_get_reading_order(ro, elem)
 
-# same as polygon_for_parent pattern in other processors
-def ensure_consistent(child):
-    """Clip segment element polygon to parent polygon range."""
-    points = child.get_Coords().points
-    polygon = polygon_from_points(points)
-    parent = child.parent_object_
-    childp = Polygon(polygon)
-    if isinstance(parent, PageType):
-        if parent.get_Border():
-            parentp = Polygon(polygon_from_points(parent.get_Border().get_Coords().points))
-        else:
-            parentp = Polygon([[0, 0], [0, parent.get_imageHeight()],
-                               [parent.get_imageWidth(), parent.get_imageHeight()],
-                               [parent.get_imageWidth(), 0]])
+def shrink_regions(page_image, page_coords, page, page_id, padding=0):
+    """Shrink each region outline to become the minimal concave hull of its binary foreground."""
+    LOG = getLogger('processor.RepairSegmentation')
+    page_array = ~ np.array(page_image.convert('1'))
+    page_polygon = page_poly(page)
+    if page.get_Border():
+        page_polygon = Polygon(polygon_from_points(page.get_Border().get_Coords().points))
+    # estimate glyph scale (roughly)
+    _, components = cv2.connectedComponents(page_array.astype(np.uint8))
+    _, counts = np.unique(components, return_counts=True)
+    if counts.shape[0] > 1:
+        counts = np.sqrt(3 * counts)
+        counts = counts[(5 < counts) & (counts < 100)]
+        scale = int(np.median(counts))
     else:
-        parentp = Polygon(polygon_from_points(parent.get_Coords().points))
+        scale = 43
+    for region in page.get_AllRegions():
+        #LOG.info('Shrinking region "%s"', region.id)
+        region_mask = np.zeros_like(page_array, dtype=np.bool)
+        region_polygon = coordinates_of_segment(region, page_image, page_coords)
+        region_mask[draw.polygon(region_polygon[:, 1],
+                                 region_polygon[:, 0],
+                                 region_mask.shape)] = True
+        # find outer contour (parts):
+        region_array = page_array & region_mask
+        contours, _ = cv2.findContours(region_array.astype(np.uint8),
+                                       cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
+        # determine areas of parts:
+        areas = [cv2.contourArea(contour) for contour in contours]
+        total_area = sum(areas)
+        if not total_area:
+            # ignore if too small
+            LOG.warning('Zero contour area in region "%s"', region.id)
+            continue
+        # pick contour and convert to absolute:
+        region_polygon = join_polygons([make_valid(Polygon(contour[:, 0, ::]))
+                                        for contour in contours
+                                        if len(contour) >= 3], loc=region.id, scale=scale)
+        if padding:
+            region_polygon = region_polygon.buffer(padding)
+        region_polygon = coordinates_for_segment(region_polygon.exterior.coords[:-1], page_image, page_coords)
+        region_polygon = Polygon(region_polygon)
+        if not region_polygon.within(page_polygon):
+            region_polygon = clip_poly(region_polygon, page_polygon)
+        if region_polygon is not None:
+            LOG.debug('Using new coordinates for region "%s"', region.id)
+            region.get_Coords().set_points(points_from_polygon(region_polygon.exterior.coords[:-1]))
+
+def simplify(segment, tolerance=0):
+    if tolerance <= 0:
+        return # nothing to do
+    coords = segment.get_Coords()
+    polygon = polygon_from_points(coords.points)
+    if len(polygon) < 10:
+        return # already simple enough
+    poly = Polygon(polygon)
+    if poly.is_empty:
+        return
+    if poly.length < tolerance * 4:
+        return # already too small
+    if poly.length > tolerance * len(polygon):
+        return # already fewer points
+    if np.sqrt(poly.area) < 0.1 * tolerance:
+        return # already relatively small
+    # simplify may easily clip out foreground
+    #poly = poly.simplify(tolerance)
+    # get equidistant list of points along hull
+    poly = Polygon([poly.exterior.interpolate(dist).coords[0] # .xy
+                    for dist in np.arange(0, poly.length, tolerance)])
+    if poly.minimum_clearance < 1.0:
+        # follow-up calculations will necessarily be integer;
+        # so anticipate rounding here and then ensure validity
+        poly = Polygon(np.round(poly.exterior.coords))
+    poly = make_valid(poly)
+    polygon = poly.exterior.coords[:-1]
+    coords.set_points(points_from_polygon(polygon))
+
+def merge_poly(poly1, poly2):
+    poly = poly1.union(poly2)
+    if poly.type == 'MultiPolygon':
+        #poly = poly.convex_hull
+        poly = join_polygons(poly.geoms)
+    if poly.minimum_clearance < 1.0:
+        poly = Polygon(np.round(poly.exterior.coords))
+    poly = make_valid(poly)
+    return poly
+
+def clip_poly(poly1, poly2):
+    poly = poly1.intersection(poly2)
+    if poly.is_empty or poly.area == 0.0:
+        return None
+    if poly.type == 'GeometryCollection':
+        # heterogeneous result: filter zero-area shapes (LineString, Point)
+        poly = unary_union([geom for geom in poly.geoms if geom.area > 0])
+    if poly.type == 'MultiPolygon':
+        # homogeneous result: construct convex hull to connect
+        #poly = poly.convex_hull
+        poly = join_polygons(poly.geoms)
+    if poly.minimum_clearance < 1.0:
+        # follow-up calculations will necessarily be integer;
+        # so anticipate rounding here and then ensure validity
+        poly = Polygon(np.round(poly.exterior.coords))
+        poly = make_valid(poly)
+    return poly
+    
+def page_poly(page):
+    return Polygon([[0, 0],
+                    [0, page.get_imageHeight()],
+                    [page.get_imageWidth(), page.get_imageHeight()],
+                    [page.get_imageWidth(), 0]])
+
+# same as polygon_for_parent pattern in other processors
+def ensure_consistent(child, at_parent=False):
+    """Make segment coordinates fit into parent coordinates.
+    
+    Ensure that the coordinate polygon of ``child`` is fully
+    contained in the coordinate polygon of its parent.
+    
+    \b
+    To achieve that when necessary, either
+    - enlarge the parent to the union of both,
+      if ``at_parent``
+    - shrink the child to the intersection of both,
+      otherwise.
+    
+    In any case, ensure the resulting polygon is valid.
+    
+    If the parent is at page level, and there is no Border,
+    then use the page frame (and assume `at_parent=False`).
+    
+    If ``child`` is at page level, and there is a Border,
+    then use the page frame as parent (and assume `at_parent=False`).
+    """
+    if isinstance(child, PageType):
+        if not child.get_Border():
+            return
+        childp = Polygon(polygon_from_points(child.get_Border().get_Coords().points))
+        parentp = page_poly(child)
+        at_parent = False # clip to page frame
+        parent = child
+    else:
+        points = child.get_Coords().points
+        polygon = polygon_from_points(points)
+        parent = child.parent_object_
+        childp = Polygon(polygon)
+        if isinstance(parent, PageType):
+            if parent.get_Border():
+                parentp = Polygon(polygon_from_points(parent.get_Border().get_Coords().points))
+            else:
+                parentp = page_poly(parent)
+                at_parent = False # clip to page frame
+        else:
+            parentp = Polygon(polygon_from_points(parent.get_Coords().points))
     # ensure input coords have valid paths (without self-intersection)
     # (this can happen when shapes valid in floating point are rounded)
     childp = make_valid(childp)
     parentp = make_valid(parentp)
-    # check if clipping is necessary
     if childp.within(parentp):
         return
-    # clip to parent
-    interp = childp.intersection(parentp)
-    if interp.is_empty or interp.area == 0.0:
-        raise Exception("Segment '%s' does not intersect its parent '%s'" % (
-            child.id, parent.id))
-    if interp.type == 'GeometryCollection':
-        # heterogeneous result: filter zero-area shapes (LineString, Point)
-        interp = unary_union([geom for geom in interp.geoms if geom.area > 0])
-    if interp.type == 'MultiPolygon':
-        # homogeneous result: construct convex hull to connect
-        # FIXME: construct concave hull / alpha shape
-        interp = interp.convex_hull
-    if interp.minimum_clearance < 1.0:
-        # follow-up calculations will necessarily be integer;
-        # so anticipate rounding here and then ensure validity
-        interp = asPolygon(np.round(interp.exterior.coords))
-        interp = make_valid(interp)
-    polygon = interp.exterior.coords[:-1] # keep open
-    points = points_from_polygon(polygon)
-    child.get_Coords().set_points(points)
+    # enlargement/clipping is necessary
+    if at_parent:
+        # enlarge at parent
+        unionp = merge_poly(childp, parentp)
+        polygon = unionp.exterior.coords[:-1] # keep open
+        points = points_from_polygon(polygon)
+        parent.get_Coords().set_points(points)
+    else:
+        # clip to parent
+        interp = clip_poly(childp, parentp)
+        if interp is None:
+            raise Exception("Segment '%s' does not intersect its parent '%s'" % (
+                child.id, parent.id))
+        polygon = interp.exterior.coords[:-1] # keep open
+        points = points_from_polygon(polygon)
+        child.get_Coords().set_points(points)
 
 def ensure_valid(element):
     changed = False
